class Tab(object): pass
class ChatTab(Tab): pass
class PrivateTab(ChatTab): pass
class MucTab(ChatTab): pass
class ConversationTab(ChatTab): pass
class RosterInfoTab(Tab): pass
class XMLTab(Tab): pass
class DynamicConversationTab(Tab): pass
class StaticConversationTab(Tab): pass
class GapTab(Tab): pass

